# -*- coding: utf-8 -*-
"""
Library for efficient processing and visualization of time series.
"""
from .ts import TimeSeries
from .tsdb import TsDB
from ._version import __version__, __version_tuple__

# summary
__summary__ = __doc__
